import torch
import torch.nn.functional as F
from einops import rearrange, repeat

# Pure-PyTorch reference selective scan (no CUDA/Triton deps)
def selective_scan_fn(u, delta, A, B, C, D=None, z=None, delta_bias=None, delta_softplus=False,
                     return_last_state=False):
    return selective_scan_ref(u, delta, A, B, C, D, z, delta_bias, delta_softplus, return_last_state)


def selective_scan_ref(u, delta, A, B, C, D=None, z=None, delta_bias=None, delta_softplus=False,
                      return_last_state=False):
    dtype_in = u.dtype
    u = u.float()
    delta = delta.float()
    if delta_bias is not None:
        delta = delta + delta_bias[..., None].float()
    if delta_softplus:
        delta = F.softplus(delta)
    batch, dim, dstate = u.shape[0], A.shape[0], A.shape[1]
    is_variable_B = B.dim() >= 3
    is_variable_C = C.dim() >= 3
    if A.is_complex():
        if is_variable_B:
            B = torch.view_as_complex(rearrange(B.float(), "... (L two) -> ... L two", two=2))
        if is_variable_C:
            C = torch.view_as_complex(rearrange(C.float(), "... (L two) -> ... L two", two=2))
    else:
        B = B.float()
        C = C.float()
    x = A.new_zeros((batch, dim, dstate))
    ys = []
    deltaA = torch.exp(torch.einsum('bdl,dn->bdln', delta, A))
    if not is_variable_B:
        deltaB_u = torch.einsum('bdl,dn,bdl->bdln', delta, B, u)
    else:
        if B.dim() == 3:
            deltaB_u = torch.einsum('bdl,bnl,bdl->bdln', delta, B, u)
        else:
            B = repeat(B, "B G N L -> B (G H) N L", H=dim // B.shape[1])
            deltaB_u = torch.einsum('bdl,bdnl,bdl->bdln', delta, B, u)
    if is_variable_C and C.dim() == 4:
        C = repeat(C, "B G N L -> B (G H) N L", H=dim // C.shape[1])
    last_state = None
    for i in range(u.shape[2]):
        x = deltaA[:, :, i] * x + deltaB_u[:, :, i]
        if not is_variable_C:
            y = torch.einsum('bdn,dn->bd', x, C)
        else:
            if C.dim() == 3:
                y = torch.einsum('bdn,bn->bd', x, C[:, :, i])
            else:
                y = torch.einsum('bdn,bdn->bd', x, C[:, :, :, i])
        if i == u.shape[2] - 1:
            last_state = x
        if y.is_complex():
            y = y.real * 2
        ys.append(y)
    y = torch.stack(ys, dim=2) # (batch dim L)
    out = y if D is None else y + u * rearrange(D, "d -> d 1")
    if z is not None:
        out = out * F.silu(z)
    out = out.to(dtype=dtype_in)
    if return_last_state:
        return out, last_state
    return out



import torch
import torch.nn.functional as F


def _apply_activation(x, activation):
    if activation is None or activation == "identity":
        return x
    if activation in ("silu", "swish"):
        return F.silu(x)
    if activation == "relu":
        return F.relu(x)
    raise ValueError(f"Unsupported activation: {activation}")


def causal_dilated_conv1d_fn(x, weight, bias=None, activation=None, dilation=1):
    """
    Depthwise causal 1D convolution with dilation.
    x: (B, D, L)
    weight: (D, W)
    bias: (D,) or None
    """
    if dilation < 1:
        raise ValueError(f"dilation must be >= 1, got {dilation}")
    pad = dilation * (weight.shape[-1] - 1)
    x = F.pad(x, (pad, 0))
    y = F.conv1d(
        x,
        weight.unsqueeze(1),
        bias=bias,
        stride=1,
        padding=0,
        dilation=dilation,
        groups=weight.shape[0],
    )
    return _apply_activation(y, activation)


def causal_dilated_conv1d_update(x, conv_state, weight, bias=None, activation=None, dilation=1):
    """
    Single-step causal dilated conv update.
    x: (B, D)
    conv_state: (B, D, S) where S = dilation * (W - 1) + 1
    weight: (D, W)
    bias: (D,) or None
    """
    if dilation < 1:
        raise ValueError(f"dilation must be >= 1, got {dilation}")
    conv_state.copy_(torch.roll(conv_state, shifts=-1, dims=-1))
    conv_state[:, :, -1] = x

    # Align taps so weight[0] multiplies the oldest sample and weight[-1] the newest.
    idx = torch.arange(weight.shape[-1] - 1, -1, -1, device=conv_state.device) * dilation
    pos = conv_state.shape[-1] - 1 - idx
    values = conv_state.index_select(-1, pos)
    y = torch.sum(values * weight.unsqueeze(0), dim=-1)
    if bias is not None:
        y = y + bias
    return _apply_activation(y, activation)
